#include "ds50daq/DAQ/detail/V1190_io.hh"
#include <unistd.h>
#include "cetlib/exception.h"

using namespace detail;

V1190_io::V1190_io (uint32_t vme_base, const std::string& link_type, int link_number, int link_node): irq_level_(0) {
  
  caen_io_ = std::unique_ptr<caen_io>(new caen_io(cvA32_U_DATA, cvD16, vme_base, link_type, link_number, link_node));

  reset_ ();

  if (caen_io_->read_register (V1190::oui0) != 0xE6 || caen_io_->read_register (V1190::oui1) != 0x40 || caen_io_->read_register (V1190::oui2) != 0)
    throw cet::exception("V1190_io") << "module not recognised";

  caen_io_->set_register_bits (V1190::control, V1190::compensation_enable_msk | V1190::berr_enable_msk | V1190::align64_msk);
  caen_io_->write_register (V1190::blt_event_num, 1);

  enable_irq_ ();
}

void V1190_io::clear () { 
  caen_io_->write_register (V1190::sw_clear, 0); 
  usleep (300000);
  wait_uc_ (V1190::write_ok_msk, 1000); 
}
  
bool V1190_io::wait_irq (uint32_t timeout_ms) {
  if (!irq_level_) throw cet::exception("V1190_io") << "irq not enabled";
  return caen_io_->wait_irq (CVIRQLevels(irq_level_), timeout_ms);
}

void V1190_io::soft_trigger () { caen_io_->write_register (V1190::sw_trigger, 0); }

void V1190_io::geo (int id) { caen_io_->write_register (V1190::geo_address, (id & 0x1F)); }

uint16_t V1190_io::control () { return caen_io_->read_register (V1190::control); }
void V1190_io::empty_events (bool on) {
  if (on) caen_io_->set_register_bits (V1190::control, V1190::empty_event_msk);
  else caen_io_->clear_register_bits (V1190::control, V1190::empty_event_msk); 
}
void V1190_io::termination (bool on) {
  if (on) caen_io_->set_register_bits (V1190::control, V1190::term_msk | V1190::term_sw_msk);
  else caen_io_->clear_register_bits (V1190::control, V1190::term_msk | V1190::term_sw_msk);
}
void V1190_io::trigger_time_tag (bool on) {
  if (on) caen_io_->set_register_bits (V1190::control, V1190::trigger_time_tag_enable_msk);
  else caen_io_->clear_register_bits (V1190::control, V1190::trigger_time_tag_enable_msk); 
}

uint16_t V1190_io::status () { return caen_io_->read_register (V1190::status); }

uint16_t V1190_io::events_stored () { return caen_io_->read_register (V1190::event_stored); }

uint16_t V1190_io::events_counter () { return caen_io_->read_register (V1190::event_counter_d32, cvD32); }

void V1190_io::edge_resolution (V1190::edge_resolutions r) { send_uc_ (V1190::set_tr_lead_lsb, r); }

void V1190_io::acquisition_mode (V1190::acquisition_modes mode) { write_uc_ ((mode == V1190::trigger_matching) ? V1190::trg_match : V1190::cont_store); }
V1190::acquisition_modes V1190_io::acquisition_mode () { return (recv_uc_ (V1190::read_acq_mod) & 1) ? V1190::trigger_matching : V1190::continous_storage; }

void V1190_io::edge_detection (V1190::edge_detections mode) { send_uc_ (V1190::set_detection, mode); }
V1190::edge_detections V1190_io::edge_detection () { 
  switch (recv_uc_ (V1190::read_acq_mod) & 3) {
    case 0:
      return V1190::pair_mode;
    case 1:
      return V1190::trailing_only;
    case 2:
      return V1190::leading_only;
    default:
      return V1190::trailing_and_leading;
  }
}
void V1190_io::matching_window_width_25ns (u_int16_t width) { 
  if (width > 0xFFF) throw cet::exception("V1190_io") << "matching_window_width_25ns excede limit of 4095 " << width;
  send_uc_ (V1190::set_win_width, width); 
}
void V1190_io::matching_window_offset_25ns (int16_t offset) { 
  if (offset > 0x28 && offset < -0x800) throw cet::exception("V1190_io") << "matching_window_offset_25ns excede limit of (-2048,40) " << offset;
  send_uc_ (V1190::set_win_offset, offset); 
}
void V1190_io::tdc_header (bool on) { write_uc_ (on ? V1190::en_head_trailer : V1190::dis_head_trailer); }
void V1190_io::trigget_time_subtraction (bool on) { write_uc_ (on ? V1190::en_sub_trg : V1190::dis_sub_trg); }

ssize_t V1190_io::read_event (void* destination, size_t max_size) {
  if (!data_ready ()) return 0;
  return caen_io_->read_block (V1190::output, destination, max_size, false, cvA32_U_BLT);
}

void V1190_io::wait_uc_ (V1190::micro_codes_msk mode, int timeout_cycles) {
  for (int i = 0; i < timeout_cycles; i++) {
    if (caen_io_->read_register (V1190::micro_hnd) & mode) return;
    usleep (100000);
  }

  throw cet::exception("V1190_io") << "wait_mc timeout for " << mode;
}

void V1190_io::write_uc_ (u_int16_t data) {
  wait_uc_ (V1190::write_ok_msk);
  usleep (100);
  caen_io_->write_register (V1190::micro, data);
}

u_int16_t V1190_io::read_uc_ () {
  wait_uc_ (V1190::read_ok_msk);
  usleep (100);
  return caen_io_->read_register (V1190::micro);
}

void V1190_io::send_uc_ (V1190::uc_opcodes oc, u_int16_t data) {
  write_uc_ (oc);
  write_uc_ (data);
}

u_int16_t V1190_io::recv_uc_ (V1190::uc_opcodes oc) {
  write_uc_ (oc);
  return read_uc_ ();
}

void V1190_io::reset_ () { 
  caen_io_->write_register (V1190::mod_reset, 0); 
  usleep (300000);
  wait_uc_ (V1190::write_ok_msk, 1000); 
}

void V1190_io::enable_irq_ (uint8_t level, uint8_t vector) {
  irq_level_ = level;
  caen_io_->write_register (V1190::int_level, level);
  caen_io_->write_register (V1190::int_vector, vector);
  caen_io_->write_register (V1190::almost_full_lvl, 5);
}
