#include "caen_io.hh"
#ifndef LINUX
#define LINUX
#include "CAENVMElib.h"
#undef LINUX
#else
#include "CAENVMElib.h"
#endif

#ifdef _IO_DEBUG_
#include <iostream>
#endif
#include <map>
#include "cetlib/exception.h"

#include "const_map_accessors.hh"

namespace cma = detail::const_map_accessors;


namespace {
  const std::string caen_error (CVErrorCodes err) {
    static const std::map<CVErrorCodes, std::string> errors{{cvSuccess, "cvSuccess" }, 
    						{ cvBusError, "cvBusError" },
					   	{ cvCommError, "cvCommError" },
					     	{ cvGenericError, "cvGenericError" },
					     	{ cvInvalidParam, "cvInvalidParam" },
					     	{ cvTimeoutError, "cvTimeoutError" }};
    auto it = errors.find (err);
    if (it == errors.end ()) return "cvUnknownError";
    return it->second;
  }


   static const std::map<std::string, CVBoardTypes> link_types{{ "USB", cvV1718 },
     								{ "PCI", cvV2718 },
     								{ "PCIE", cvV2718 }};

}

detail::caen_io::caen_io (CVAddressModifier am, CVDataWidth dm, uint32_t vme_base, const std::string& link_type, int link_number, int link_node):
	am_(am), dm_(dm), vme_base_(vme_base) {
  switch (dm_) {
    case cvD32:
      mask_ = 0xffffffffUL; 
      b4b_ = 4;
      break;
    case cvD16:
      mask_ = 0xffff;
      b4b_ = 2;
      break;
    case cvD8:
      mask_ = 0xff;
      b4b_ = 1;
      break;
    default:
      cet::exception("caen_io") << "unsupported data width " << dm_;
  };

  CVBoardTypes link_type_ = cma::get (link_types, link_type, "link_type");
  CVErrorCodes err = CAENVME_Init (link_type_, link_number, link_node, &handle_);
  if (err != cvSuccess) throw cet::exception("caen_io") << "CAENVME_Init(" << cma::rget(link_types, link_type_) << "," << link_number << "," << link_node << "," << std::hex << std::showbase << vme_base << "): " << caen_error (err);

#ifdef _IO_DEBUG_
  debug_ = false;
#endif
}


detail::caen_io::~caen_io () {
  CVErrorCodes err = CAENVME_End (handle_);
  if (err != cvSuccess) throw cet::exception("caen_io") << "CAENVME_End(" << handle_ << "): " << caen_error (err);
}

void detail::caen_io::write_register (uint32_t offset, uint32_t value, CVDataWidth dm) {
  CVErrorCodes err = CAENVME_WriteCycle (handle_, vme_base_ + offset, &value, am_, dm);
  if (err != cvSuccess) throw cet::exception("caen_io") << "CAENVME_WriteCycle(" <<  handle_ << "," << std::hex << std::showbase << vme_base_ + offset << "," << value << "): " << caen_error (err);
#ifdef _IO_DEBUG_
  if (debug_) std::cout << std::hex << std::showbase << offset << " <- " << value << std::endl;
#endif
}

uint32_t detail::caen_io::set_register_bits (uint32_t offset, uint32_t mask, CVDataWidth dm) {
  uint32_t v = read_register (offset, dm);
  
  if (mask) {
    v |= mask;
    write_register (offset, v & mask_, dm);
  }

  return v & mask_;
}

uint32_t detail::caen_io::clear_register_bits (uint32_t offset, uint32_t mask, CVDataWidth dm) {
  uint32_t v = read_register (offset, dm);

  if (mask) {
    v &= ~mask;
    write_register (offset, v & mask_, dm);
  }

  return v & mask_;
}

uint32_t detail::caen_io::read_register (uint32_t offset, CVDataWidth dm) {
  uint32_t v;
  CVErrorCodes err = CAENVME_ReadCycle (handle_, vme_base_ + offset, &v, am_, dm);
  if (err != cvSuccess) throw cet::exception("caen_io") << "CAENVME_ReadCycle(" <<  handle_ << "," << std::hex << std::showbase << vme_base_ + offset << "): " << caen_error (err);

#ifdef _IO_DEBUG_
  if (debug_) std::cout << std::hex << std::showbase << offset << " -> " << v << std::endl;
#endif
  return v & mask_;
}

ssize_t detail::caen_io::read_block (uint32_t offset, void* destination, size_t words, bool fifo) {
  switch (am_) {
    case cvA32_U_DATA:
      return read_block (offset, destination, words, fifo, cvA32_U_BLT);
    case cvA24_U_DATA:
      return read_block (offset, destination, words, fifo, cvA24_U_BLT);
    default:
      throw cet::exception("caen_io") << "read_block BLT not possible unless AM is specified";
  };
  return 0;
}

ssize_t detail::caen_io::read_block (uint32_t offset, void* destination, size_t words, bool fifo, CVAddressModifier am) {
  int rbytes, abytes = words * b4b_; // CAENVME_FIFOBLTReadCycle expect an int
  if (fifo) {
    CVErrorCodes err = CAENVME_FIFOBLTReadCycle (handle_, vme_base_ + offset, destination, abytes, am, cvD32, &rbytes);
    if (err == cvBusError) rbytes = -rbytes;
    else if (err != cvSuccess) throw cet::exception("caen_io") << "CAENVME_FIFOBLTReadCycle(" <<  handle_ << "," << std::hex << std::showbase << vme_base_ + offset << "," << std::dec << abytes << "): " << caen_error (err);
  } else {
    CVErrorCodes err = CAENVME_BLTReadCycle (handle_, vme_base_ + offset, destination, abytes, am, cvD32, &rbytes);
    if (err == cvBusError) rbytes = -rbytes;
    else if (err != cvSuccess) throw cet::exception("caen_io") << "CAENVME_BLTReadCycle(" <<  handle_ << "," << std::hex << std::showbase << vme_base_ + offset << "," << std::dec << abytes << "): " << caen_error (err);
  }
#ifdef _IO_DEBUG_
  if (debug_) std::cout << std::hex << std::showbase << offset << " --> " << std::dec << rbytes << std::endl;
#endif
  return rbytes;
}


void detail::caen_io::set_outputs () {
  CVErrorCodes err = CAENVME_SetOutputRegister(handle_, cvOut0Bit | cvOut1Bit | cvOut2Bit | cvOut3Bit | cvOut4Bit);
  if (err != cvSuccess) throw cet::exception("caen_io") << "CAENVME_SetOutputRegister(" << handle_ << ",cvOut0Bit | cvOut1Bit | cvOut2Bit | cvOut3Bit | cvOut4Bit): " << caen_error (err);
}
 
void detail::caen_io::clear_outputs () {
  CVErrorCodes err = CAENVME_ClearOutputRegister(handle_, 0xFFFF);
  if (err != cvSuccess) throw cet::exception("caen_io") << "CAENVME_ClearOutputRegister(" << handle_ << ",0xFFFF): " << caen_error (err);
}

void detail::caen_io::enable_outputs () {
#define CAENVME_SETOUTPUTCONF(i) do { \
  CVErrorCodes err = CAENVME_SetOutputConf (handle_, cvOutput ## i, cvDirect, cvActiveHigh, cvManualSW); \
  if (err != cvSuccess) throw cet::exception("caen_io") << "CAENVME_SetOutputConf(" << handle_ << "," << i << "cvDirect, cvActiveHigh, cvManualSW):" << caen_error (err); \
  } while (0)

  CAENVME_SETOUTPUTCONF(0);
  CAENVME_SETOUTPUTCONF(1);
  CAENVME_SETOUTPUTCONF(2);
  CAENVME_SETOUTPUTCONF(3);
  CAENVME_SETOUTPUTCONF(4);

#undef CAENVME_SETOUTPUTCONF
}


bool detail::caen_io::wait_irq (CVIRQLevels irq, uint32_t timeout_ms) {
  CVErrorCodes err = CAENVME_IRQEnable (handle_, irq);
  if (err != cvSuccess) throw cet::exception("caen_io") << "CAENVME_IRQEnable(" << handle_ << "," << std::hex << std::showbase << irq << "): " << caen_error (err);

  bool ret = true;
  err = CAENVME_IRQWait (handle_, irq, timeout_ms);
  if (err == cvTimeoutError) ret = false;
  else if (err != cvSuccess) throw cet::exception("caen_io") << "CAENVME_IRQWait(" << handle_ << "," << std::hex << std::showbase << irq << "," << timeout_ms << "): " << caen_error (err);

  err = CAENVME_IRQDisable (handle_, irq);
  if (err != cvSuccess) throw cet::exception("caen_io") << "CAENVME_IRQDisable(" << handle_ << "," << std::hex << std::showbase << irq << "): " << caen_error (err);

  return ret;
}

bool detail::caen_io::wait_irq (CVIRQLevels irq, uint32_t timeout_ms, uint32_t& vector, CVDataWidth dm) {
  if (!wait_irq (irq, timeout_ms)) return false;

  CVErrorCodes err = CAENVME_IACKCycle (handle_, irq, &vector, dm);
  if (err != cvSuccess) throw cet::exception("caen_io") << "CAENVME_IACKCycle(" << handle_ << "," << std::hex << std::showbase << irq << "): " << caen_error (err);
  return true;
}

void detail::caen_io::vme_sysreset () {
  CVErrorCodes err = CAENVME_SystemReset (handle_);
  if (err != cvSuccess) throw cet::exception("caen_io") << "CAENVME_SystemReset(" << handle_ << "): " << caen_error (err);
}

