#!/bin/bash

source `which setPortNumbers.sh`

# this function expects a number of arguments:
#  1) the DAQ command to be sent
#  2) the run number (dummy for non-start commands)
#  3) the pulser trigger rate
#  4) the laser trigger rate
#  5) whether to enable the random trigger (0 or 1)
#  6) the acqusition gate width (usec)
#  7) the majority logic low threshold
#  8) the compression level for ADC data [0..2]
#  9) whether to run online monitoring
# 10) the data directory
# 11) the logfile name
# 12) whether to write data to disk [0,1]
function launch() {
  ebComp=$8
  agComp=$8
  if [[ "$8" == "2" ]]; then
    ebComp=1
    agComp=3
  elif [[ "$8" == "1" ]]; then
    ebComp=1
    agComp=0
  fi

  ds50MasterControl.rb -s -c $1 \
    --v1495 dsfr6,$frPort0,0,$3,$4,$5,$6,$7,1 \
    --v1720 dsfr6,$frPort1,1,0,$6 \
    --v1720 dsfr6,$frPort2,2,0,$6 \
    --eb dseb8,$ebPort0,$ebComp \
    --eb dseb8,$ebPort1,$ebComp \
    --eb dseb8,$ebPort2,$ebComp \
    --eb dseb8,$ebPort3,$ebComp \
    --eb dseb8,$ebPort4,$ebComp \
    --ag dsfr6,$agPort0,3,$agComp \
    --data-dir ${10} --online-monitoring $9 \
    --write-data ${12} \
    --run-number $2 2>&1 | tee -a ${11}
}

scriptName=`basename $0`
usage () {
    echo "
Usage: ${scriptName} [options] <command>
Where command is one of:
  init, start, pause, resume, stop, or shutdown
Options:
  -h, --help: prints this usage message
  -p <pulser rate>: specifies the rate of the pulser trigger
  -l <laser rate>: specifies the rate of the laser trigger
  -r : enables the random trigger
  -g <gate width>: specifies the acquisition gate width in usec [default=400]
  -t <low threshold>: specifies the low threshold for the majority
      logic trigger [default=5]
  -c <compression level>: specifies the ADC data compression level
      0 = no compression
      1 = compression, both raw and compressed data kept
      2 = compression, only compressed data kept [default]
  -m <on|off>: specifies whether to run online monitoring [default=off]
  -o <data dir>: specifies the directory for data files [default=/data/test]
  -C <comment>: specify the run comment
  -T <type>: specify the run type
  -N <run number>: specifies the run number
  -D : disables the writing of data to disk
Notes:
  the start command expects a run number to be specified (-N).
Example: ${scriptName} -p 32768 init
         ${scriptName} -N 101 start
" >&2
}

# parse the command-line options
originalCommand="$0 $*"
pulserRate=""
laserRate=""
randomTrigger=""
gateWidth="400"
lowThreshold=5
compressionLevel=2
onmonEnable=off
diskWriting=1
dataDir="/data/test"
runNumber=""
runComment=""
runType=""
OPTIND=1
while getopts "hp:l:rg:c:N:C:T:o:t:m:D-:" opt; do
    if [ "$opt" = "-" ]; then
        opt=$OPTARG
    fi
    case $opt in
        h | help)
            usage
            exit 1
            ;;
        p)
            pulserRate=${OPTARG}
            ;;
        l)
            laserRate=${OPTARG}
            ;;
        r)
            randomTrigger=1
            ;;
        g)
            gateWidth=${OPTARG}
            ;;
        t)
            lowThreshold=${OPTARG}
            ;;
        c)
            compressionLevel=${OPTARG}
            ;;
        N)
            runNumber=${OPTARG}
            ;;
        C)
            runComment=${OPTARG}
            ;;
        T)
            runType=${OPTARG}
            ;;
        m)
            onmonEnable=${OPTARG}
            ;;
        o)
            dataDir=${OPTARG}
            ;;
        D)
            diskWriting=0
            ;;
        *)
            usage
            exit 1
            ;;
    esac
done
shift $(($OPTIND - 1))

# fetch the command to run
if [ $# -lt 1 ]; then
    usage
    exit
fi
command=$1
shift

# verify that the command is one that we expect
if [[ "$command" != "init" ]] && \
   [[ "$command" != "start" ]] && \
   [[ "$command" != "pause" ]] && \
   [[ "$command" != "resume" ]] && \
   [[ "$command" != "stop" ]] && \
   [[ "$command" != "shutdown" ]]; then
    echo "Invalid command."
    usage
    exit
fi

# verify that the expected arguments were provided
if [[ "$command" == "start" ]] && [[ "$runNumber" == "" ]]; then
    echo ""
    echo "*** A run number needs to be specified."
    usage
    exit
fi

# fill in values for options that weren't specified
if [[ "$pulserRate" == "" ]]; then
    pulserRate=0
fi
if [[ "$laserRate" == "" ]]; then
    laserRate=0
fi
if [[ "$randomTrigger" == "" ]]; then
    randomTrigger=0
fi
if [[ "$runNumber" == "" ]]; then
    runNumber=101
fi

# translate the onmon enable flag
if [[ "$onmonEnable" == "on" ]]; then
    onmonEnable=1
else
    onmonEnable=0
fi

# verify that we have online monitoring, disk writing, or both
#if [[ "$command" == "init" ]]; then
#    if [[ "$onmonEnable" == "0" ]] && [[ "$diskWriting" == "0" ]]; then
#        echo ""
#        echo "*** One or both of disk writing or online monitoring must be enabled."
#        usage
#        exit
#    fi
#fi

# build the logfile name
TIMESTAMP=`date '+%Y%m%d%H%M%S'`
logFile="/data/logs/masterControl/dsMC-${TIMESTAMP}-${command}.log"
echo "${originalCommand}" > $logFile

# invoke the requested command
launch $command $runNumber $pulserRate $laserRate $randomTrigger \
    $gateWidth $lowThreshold $compressionLevel $onmonEnable $dataDir \
    $logFile $diskWriting
